from pathlib import Path
from types import SimpleNamespace

import typer
from openai import AuthenticationError, NotFoundError, OpenAI
from typing_extensions import Annotated

from .diff import Diff
from .errors import (
    DirtyRepo,
    InvalidRemote,
    InvalidTree,
    IsAncestor,
    NotAncestor,
    NotARepo,
)
from .utils import handle_create_patch_errors, read_prompt, stream_to_console


def print_patch_review(ctx: typer.Context):
    try:
        review = ctx.obj.diff.review_patch(
            ctx.obj.client, ctx.obj.model, ctx.obj.temperature, ctx.obj.system_prompt
        )
    except Exception as err:
        print("Error while reviewing the code changes.")
        raise typer.Abort() from err
    stream_to_console(review)


app = typer.Typer()

@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    openai_api_key: Annotated[
        str,
        typer.Option(
            help="OpenAI API key", envvar="OPENAI_API_KEY", rich_help_panel="OpenAI Parameters"
        ),
    ],
    model: Annotated[
        str, typer.Option(help="OpenAI GPT model", rich_help_panel="OpenAI Parameters")
    ] = "gpt-4-turbo-preview",
    temperature: Annotated[
        int,
        typer.Option(
            min=0, max=2, help="Temperature for the model", rich_help_panel="OpenAI Parameters"
        ),
    ] = 1,
    system_prompt: Annotated[
        str,
        typer.Option(
            help="Custom system prompt to use for the diff patches",
            rich_help_panel="OpenAI Parameters",
        ),
    ] = None,
    unified: Annotated[
        int,
        typer.Option(
            help="Context line length on each side of the diff hunk",
            rich_help_panel="Git Parameters",
        ),
    ] = 3,
):
    if ctx.invoked_subcommand is None:
        ctx.get_help()
    try:
        diff = Diff(Path("."), unified=unified)
    except NotARepo as not_a_repo:
        print("Current directory is not a git repository")
        raise typer.Abort() from not_a_repo

    if not model.startswith("gpt"):
        raise typer.BadParameter(
            "Only gpt models are supported", ctx=ctx, param=model, param_hint="model"
        )

    client = OpenAI(api_key=openai_api_key)
    try:
        client.models.retrieve(model=model)
    except AuthenticationError as auth_error:
        raise typer.BadParameter(
            "Invalid OpenAI API key", ctx=ctx, param=openai_api_key, param_hint="openai_api_key"
        ) from auth_error
    except NotFoundError as no_model:
        raise typer.BadParameter(
            f"{model} does not exist", ctx=ctx, param=model, param_hint="model"
        ) from no_model

    if system_prompt is None:
        system_prompt = read_prompt("default")

    ctx.obj = SimpleNamespace(
        diff=diff,
        client=client,
        openai_api_key=openai_api_key,
        model=model,
        temperature=temperature,
        system_prompt=system_prompt,
        unified=unified,
    )

@app.command()
def add(ctx: typer.Context):
    """
    Review the changes between the working tree and the index
    """
    ctx.obj.diff.add()
    handle_create_patch_errors(ctx.obj.diff)
    print_patch_review(ctx)


@app.command()
def commit(ctx: typer.Context):
    """
    Review the changes between index and the HEAD
    """
    ctx.obj.diff.commit()
    handle_create_patch_errors(ctx.obj.diff)
    print_patch_review(ctx)


@app.command()
def merge(
    ctx: typer.Context,
    feature_branch: Annotated[str, typer.Argument(help="tree to merge into the HEAD")],
):
    """
    Review the result of merging the feature branch into the HEAD
    """
    try:
        ctx.obj.diff.merge(feature_branch)
    except InvalidTree as invalid_tree:
        print(f"{feature_branch} is not a valid tree to merge into the HEAD")
        raise typer.Abort() from invalid_tree
    except IsAncestor as ancestor_tree:
        print(f"{feature_branch} is an ancestor of the HEAD, no code changes to review")
        raise typer.Abort() from ancestor_tree
    except DirtyRepo as dirty_repo:
        print(
            "Working tree has uncommitted changes, please commit or stash them to review the merge"
        )
        raise typer.Abort() from dirty_repo
    handle_create_patch_errors(ctx.obj.diff)
    print_patch_review(ctx)


@app.command()
def push(
    ctx: typer.Context, remote: Annotated[str, typer.Argument(help="remote to push to")] = "origin"
):
    """
    Review the changes between the HEAD and the remote
    """
    remote_head = f"{remote}/{ctx.obj.diff.repo.active_branch.name}"
    try:
        ctx.obj.diff.push(remote)
    except InvalidRemote as invalid_remote:
        print(f"{remote} is not a valid remote")
        raise typer.BadParameter(
            f"{remote} is not a valid remote", ctx=ctx, param=remote, param_hint="remote"
        ) from invalid_remote
    except InvalidTree as no_upstream:
        print(f"Active branch has no upstream {remote_head}, please set it to review the push")
        raise typer.Abort() from no_upstream
    except NotAncestor as not_ancestor:
        print(
            "Remote is ahead of the local branch, please pull the changes before reviewing the push"
        )
        raise typer.Abort() from not_ancestor
    handle_create_patch_errors(ctx.obj.diff)
    print_patch_review(ctx)


@app.command()
def pr(
    ctx: typer.Context,
    target_branch: Annotated[str, typer.Argument(help="target branch to compare")],
    remote: Annotated[str, typer.Argument(help="remote of the target branch")] = "origin",
):
    """
    Review the result of a pull request from HEAD to the target branch in the remote
    """
    remote_target_ref = f"{remote}/{target_branch}"
    try:
        ctx.obj.diff.pr(target_branch, remote)
    except InvalidRemote as invalid_remote:
        raise typer.BadParameter(
            f"{remote} is not a valid remote", ctx=ctx, param=remote, param_hint="remote"
        ) from invalid_remote
    except InvalidTree as invalid_tree:
        raise typer.BadParameter(
            f"{remote_target_ref} is not a valid tree",
            ctx=ctx,
            param_hint="remote/target_branch",
        ) from invalid_tree
    except IsAncestor as ancestor_tree:
        print(f"HEAD is an ancestor of {remote_target_ref}, no code changes to review")
        raise typer.Abort() from ancestor_tree
    except DirtyRepo as dirty_repo:
        print(
            "Working tree has uncommitted changes, please commit or stash them to review the pull request"  # noqa: E501
        )
        raise typer.Abort() from dirty_repo
    handle_create_patch_errors(ctx.obj.diff)
    print_patch_review(ctx)


if __name__ == "__main__":
    app()



import uuid
from pathlib import Path
from typing import List

from git import GitCommandError, InvalidGitRepositoryError, Repo, diff
from openai import OpenAI, Stream

from .errors import (
    DirtyRepo,
    InvalidRemote,
    InvalidTree,
    IsAncestor,
    NoCodeChanges,
    NoDiffs,
    NotAncestor,
    NotARepo,
)


def fetch_remote(repo: Repo, remote: str) -> None:
    """Fetch the remote.

    Args:
        repo (Repo): The git repository.
        remote (str): The remote to fetch.

    Raises:
        InvalidRemote: Raised when the remote is not found.
    """
    try:
        repo.git.fetch(remote)
    except GitCommandError as no_remote:
        raise InvalidRemote from no_remote


def check_ancestry(repo: Repo, ancestor_commit: str, commit: str = "HEAD") -> bool:
    """
    Check if the commit is an ancestor of the ancestor commit.

    Args:
        repo (Repo): Repo object to compare the commits.
        ancestor_commit (str): Commit to verify to be an ancestor.
        commit (str, optional): Commit to compare against the ancestor commit. Defaults to "HEAD".

    Raises:
        InvalidTree: If the commit or the ancestor commit is not found.

    Returns:
        bool: Whether the ancestor_commit is an ancestor of the commit.
    """
    try:
        return repo.is_ancestor(ancestor_commit, commit)
    except GitCommandError as no_tree:
        raise InvalidTree from no_tree


class Diff:
    """
    The Diff class for generating diffs and patches.
    """

    def __init__(self, repo_path: Path, unified: int = 3) -> None:
        """
        Initialize the Diff class.

        Args:
            repo_path (Path): The path to the git repository.
            unified (int, optional): The number of lines of context to include in the patch.
            Defaults to 3.

        Raises:
            NotARepo: Raised when the path is not a git repository.
        """
        try:
            self.repo = Repo(repo_path, search_parent_directories=True)
        except InvalidGitRepositoryError as no_git:
            raise NotARepo from no_git
        self.diffs = None
        self.patch = None
        self.repo_path = repo_path
        self.unified = unified

    def add(self) -> "Diff":
        """
        Set diffs to the diffs between the index and the working tree.

        Returns:
            Diff: The Diff object.
        """
        self.diffs = self.repo.index.diff(
            None, create_patch=True, no_ext_diff=True, unified=self.unified
        )
        return self

    def commit(self) -> "Diff":
        """
        Set diffs to the diffs between the HEAD and the index.

        Returns:
            Diff: The Diff object.
        """
        self.diffs = self.repo.head.commit.diff(
            create_patch=True, no_ext_diff=True, unified=self.unified
        )
        return self

    def _create_tmp_branch(self, from_commit: str = "HEAD") -> str:
        """
        Create a temporary branch from a commit.

        Args:
            from_commit (str, optional): The commit to create the branch from. Defaults to "HEAD".

        Returns:
            str: Name of the temporary branch.
        """
        tmp_branch = f"tmp-{uuid.uuid4()}"
        self.repo.create_head(tmp_branch, from_commit)
        return tmp_branch

    def _merge_on_temp_branch(self, feature_commit: str, base_commit: str) -> List[diff.Diff]:
        """
        Merge the feature commit into the base commit on a temporary branch and set the diffs.

        Args:
            feature_commit (str): The commit to merge into the base commit.
            base_commit (str): The commit to merge the feature commit into.

        Raises:
            DirtyRepo: If the repository has uncommitted changes.

        Returns:
            List[diff.Diff]: The diffs between the base and the feature commit.
        """
        if self.repo.is_dirty():
            raise DirtyRepo
        # Save current branch to get back to it later
        active_branch = self.repo.active_branch.name

        tmp_branch = self._create_tmp_branch(base_commit)
        self.repo.heads[tmp_branch].checkout()

        has_conflict = False
        try:
            self.repo.git.merge(feature_commit, no_commit=True, no_ff=True)
        except GitCommandError:
            has_conflict = True
        finally:
            diffs = self.repo.head.commit.diff(
                None, create_patch=True, no_ext_diff=True, unified=self.unified
            )
        if has_conflict:
            self.repo.git.merge(abort=True)
        else:
            self.repo.git.reset(hard=True)

        # Go back to original branch
        self.repo.heads[active_branch].checkout()
        # Clean up the temporary branch
        self.repo.delete_head(tmp_branch, force=True)

        self.diffs = diffs
        return diffs

    def merge(self, tree: str) -> "Diff":
        """
        Set diffs to the diffs between the HEAD and the tree.

        Args:
            tree (str): The tree to compare against.

        Raises:
            IsAncestor: When the tree is already an ancestor of the HEAD.

        Returns:
            Diff: The Diff object.
        """
        tree_is_ancestor = check_ancestry(self.repo, tree)
        if tree_is_ancestor:
            raise IsAncestor

        self.diffs = self._merge_on_temp_branch(tree, self.repo.active_branch.name)

        return self

    def push(self, remote: str = "origin") -> "Diff":
        """
        Diff between the HEAD and the remote HEAD.

        Args:
            remote (str, optional): The remote to compare against. Defaults to "origin".

        Raises:
            NotAncestor: When the remote HEAD is not an ancestor of the HEAD.

        Returns:
            Diff: The Diff object.
        """
        remote_head = f"{remote}/{self.repo.active_branch.name}"

        fetch_remote(self.repo, remote)
        remote_is_ancestor = check_ancestry(self.repo, remote_head)
        if not remote_is_ancestor:
            raise NotAncestor

        self.diffs = self.repo.head.commit.diff(
            remote_head, create_patch=True, no_ext_diff=True, R=True, unified=self.unified
        )
        return self

    def pr(self, target_branch: str, remote: str = "origin") -> "Diff":
        """
        Diff between the HEAD and the target branch on the remote.

        Args:
            target_branch (str): The target branch on the remote.
            remote (str, optional): The remote to compare against. Defaults to "origin".

        Raises:
            IsAncestor: When the HEAD is an ancestor of the target branch on the remote.

        Returns:
            Diff: The Diff object.
        """
        remote_head = f"{remote}/{target_branch}"
        fetch_remote(self.repo, remote)

        head_is_ancestor = check_ancestry(self.repo, "HEAD", remote_head)
        if head_is_ancestor:
            raise IsAncestor

        self.diffs = self._merge_on_temp_branch(self.repo.active_branch.name, remote_head)

        return self

    def create_patch(self) -> str:
        """
        Get the patch from the diffs.

        Raises:
            Exception: No diffs generated.
            NoDiffs: When there are no diffs to review.

        Returns:
            str: All the diffs in the patch.
        """
        if self.diffs is None:
            raise Exception("No diffs generated.")
        elif len(self.diffs) == 0:
            raise NoDiffs
        patch = "\n".join([diff.diff.decode("utf-8") for diff in self.diffs])
        if patch.strip() == "":
            raise NoCodeChanges
        self.patch = patch
        return patch

    def review_patch(
        self, openai_client: OpenAI, model: str, temperature: float, system_prompt: str
    ) -> Stream:
        """
        Review the patch using OpenAI's chat completion models.

        Args:
            openai_client (OpenAI): The OpenAI client.
            model (str): Model to use for the review.
            temperature (float): Temperature parameter for the model.
            system_prompt (str): System prompt to use for the review.

        Raises:
            Exception: When there is no patch to review.

        Returns:
            Stream: Chat completion stream.
        """
        if self.patch is None:
            raise Exception("No patch to review.")

        self.review = openai_client.chat.completions.create(
            model=model,
            messages=[
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": self.patch},
            ],
            temperature=temperature,
            stream=True,
        )

        return self.review



class NotARepo(Exception):
    "Raised when the current directory is not a git repository."

    pass


class InvalidTree(Exception):
    "Raised when the tree is not valid."

    pass

class InvalidRemote(Exception):
    "Raised when the remote is not valid."

    pass


class NoDiffs(Exception):
    "Raised when there are no diffs to review."

    pass


class NotAncestor(Exception):
    "Raised when the tree is not an ancestor of the HEAD."

    pass


class IsAncestor(Exception):
    "Raised when the tree is an ancestor of the HEAD."

    pass

class DirtyRepo(Exception):
    "Raised when the repository is dirty."

    pass


class NoCodeChanges(Exception):
    "Raised when there is a diff but no code changes."

    pass



import pkgutil

import typer
from openai import Stream

from .diff import Diff
from .errors import NoCodeChanges, NoDiffs


def stream_to_console(stream: Stream) -> str:
    """
    Prints the stream to the console and returns the full stream as a string

    Args:
        stream (Stream): Stream of text from OpenAI

    Returns:
        str: Full stream of text from OpenAI
    """
    full_stream = ""
    for chunk in stream:
        chunk_content = chunk.choices[0].delta.content
        if chunk_content is None:
            break
        full_stream += chunk_content
        print(chunk_content, end="", flush=True)
    return full_stream


def read_prompt(prompt: str) -> str:
    """
    Reads a prompt from system_prompts directory

    Args:
        prompt (str): Name of the file to read

    Returns:
        str: Contents of the file
    """
    return pkgutil.get_data(__name__, f"system_prompts/{prompt}").decode("utf-8")


def handle_create_patch_errors(diff_object: Diff) -> None:
    """
    Handles errors raised when creating a patch

    Args:
        diff_object (Diff): The Diff object
    """
    try:
        diff_object.create_patch()
    except NoDiffs as no_diffs:
        print("No differences found to review")
        raise typer.Abort() from no_diffs
    except NoCodeChanges as no_code_changes:
        print("No meaningful code changes found to review")
        raise typer.Abort() from no_code_changes
